import os
from mistralai import Mistral
from typing import List
from .utils import with_timeout

# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_mistral(messages:List[str], model_name, timeout:int=120, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to mistral format
    client = Mistral(api_key=os.environ.get("MISTRAL_API_KEY"))
    new_messages = []
    if conv_system_msg:
        new_messages.append({"role": "system", "content": conv_system_msg})
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})

    if "n" in generate_kwargs:
        generate_kwargs.pop("n") # mistral does not have n
    stream = generate_kwargs.pop("stream", False)
    @with_timeout(timeout)
    def generate_content():
        completion = client.chat.complete(
            model=model_name,
            messages=new_messages,
            **generate_kwargs,
        )
        if len(completion.choices) > 1:
            return [c.message.content for c in completion.choices]
        else:
            return completion.choices[0].message.content
    
    @with_timeout(timeout)
    def stream_content():
        completion = client.chat.stream(
            model=model_name,
            messages=new_messages,
            **generate_kwargs,
        )
        def generate_stream():
            for chunk in completion:
                if chunk.data.choices[0].delta.content is not None:
                    yield chunk.data.choices[0].delta.content
        return generate_stream()
    if not stream:
        return generate_content()
    else:
        return stream_content()
    
if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_mistral(["Hello", "Hi, I am mistral", "What did I ask in the last response?"], "mistral-large-latest"))


import os
import sys
import time
import random
import atexit
import psutil
import subprocess
from functools import partial
from .utils import generation_cache_wrapper, retry_on_failure, convert_messages_wrapper, SubprocessMonitor, MaxRetriesExceededError
from typing import Union, List
from tqdm import tqdm

import importlib.util
flash_attn = importlib.util.find_spec("flash_attn")
if not flash_attn:
    print("Warning: flash_attn not found, recommend to install flash_attn for better performance")
    print("Simple Command: pip install flash_attn --no-build-isolation")
    print("Please refer to https://github.com/Dao-AILab/flash-attention for detailed installation instructions")

ENGINES = ["vllm", "sglang", "openai", "gemini", "mistral", "together", "claude"]
all_workers = []
verbose = False
def set_verbose(value):
    global verbose
    verbose = value
    
class ModelWorker:
    def __init__(self, model_name, worker_addr, proc, gpu_ids=None):
        self.model_name = model_name
        self.worker_addr = worker_addr
        self.proc = proc
        self.gpu_ids = gpu_ids
    
    def __str__(self):
        return f"ModelWorker(model_name={self.model_name}, worker_addr={self.worker_addr}, proc={self.proc}, gpu_ids={self.gpu_ids})"
    
    def __repr__(self):
        return self.__str__()
    
def get_call_worker_func(
    model_name, 
    worker_addrs=None, 
    cache_dir=None, 
    use_cache=True,
    overwrite_cache=False,
    completion=False, 
    num_workers=1,
    num_gpu_per_worker=None,
    gpu_ids=None,
    dtype="auto",
    quantization=None,
    engine="vllm",
    additional_args=[],
    max_retry=None,
    verbose=False,
    return_worker=False,
) -> str:
    """
    Return a function that calls the model worker, takes a list of messages (user, gpt, user, ...) and returns the generated text
    Args:
        model_name: model name
        worker_addrs: worker addresses, if None, launch local workers
        cache_dir: cache directory
        use_cache: use cache or not. Cache is on the hash of the input message.
        overwrite_cache: overwrite cache or not. If True, previous cache will be overwritten.
        completion: use completion or not (use chat by default)
        num_workers: number of workers
        num_gpu_per_worker: number of gpus per worker
        dtype: data type
        engine: engine name
        additional_args: additional arguments for launching the worker (vllm, sglang)
    """
    loaded_wokers = []
    if engine == "openai":
        from .openai_text import call_worker_openai, call_worker_openai_completion
        call_model_worker = call_worker_openai if not completion else call_worker_openai_completion
    elif engine == "gemini":
        if completion:
            raise ValueError(f"Engine {engine} does not support completion")
        from .gemini import call_worker_gemini
        call_model_worker = call_worker_gemini
    elif engine == "claude":
        if completion:
            raise ValueError(f"Engine {engine} does not support completion")
        from .claude import call_worker_claude
        call_model_worker = call_worker_claude
    elif engine == "mistral":
        if completion:
            raise ValueError(f"Engine {engine} does not support completion")
        from .mistral import call_worker_mistral
        call_model_worker = call_worker_mistral
    elif engine == "together":
        from .together import call_worker_together, call_worker_together_completion
        call_model_worker = call_worker_together if not completion else call_worker_together_completion
    elif engine in ["vllm", "sglang"]:
        assert num_gpu_per_worker is not None, "num_gpu_per_worker must be provided for vllm and sglang"
        if engine == "vllm":
            from .vllm import launch_vllm_worker, call_vllm_worker, call_vllm_worker_completion
            call_worker_func = call_vllm_worker if not completion else call_vllm_worker_completion
            launch_worker_func = launch_vllm_worker
        elif engine == "sglang":
            from .sglang import launch_sglang_worker, call_sglang_worker, call_sglang_worker_completion
            call_worker_func = call_sglang_worker if not completion else call_sglang_worker_completion
            launch_worker_func = launch_sglang_worker
        else:
            raise ValueError(f"Internal error: engine {engine} not supported")
        if worker_addrs is None:
            import torch
            
            print(f"Launching model worker {model_name} locally")
            worker_addrs = []
            total_gpus = torch.cuda.device_count()
            if gpu_ids:
                gpu_ids = [int(gpu_id) for gpu_id in gpu_ids]
                assert len(gpu_ids) <= total_gpus, f"Error: number of gpus {len(gpu_ids)} is greater than total gpus {total_gpus}"
                total_gpus = len(gpu_ids)
            if total_gpus < num_workers * num_gpu_per_worker:
                if total_gpus >= num_gpu_per_worker:
                    print(f"Warning: total gpus ({total_gpus}) is less than num_workers * num_gpu_per_worker ({num_workers * num_gpu_per_worker}), using {total_gpus // num_gpu_per_worker} workers instead")
                    num_workers = total_gpus // num_gpu_per_worker
                else:
                    print(f"Error: total gpus ({total_gpus}) is less than num_gpu_per_worker ({num_gpu_per_worker}), exiting...")
                    sys.exit(1)
            if not gpu_ids:
                if os.environ.get("CUDA_VISIBLE_DEVICES") is not None:
                    gpus_ids = os.environ.get("CUDA_VISIBLE_DEVICES").split(",")
                    gpu_ids = [int(gpu_id) for gpu_id in gpus_ids]
                else:
                    gpu_ids = list(range(total_gpus))
            start_port = random.randint(31000, 32000)
            for i in range(num_workers):
                worker_addr, proc = launch_worker_func(model_name, 
                    num_gpus=num_gpu_per_worker, 
                    gpu_ids=gpu_ids[i*num_gpu_per_worker:(i+1)*num_gpu_per_worker], 
                    port=start_port+i*10,
                    dtype=dtype, quantization=quantization, additional_args=additional_args)
                worker = ModelWorker(model_name, worker_addr, proc, gpu_ids=gpu_ids[i*num_gpu_per_worker:(i+1)*num_gpu_per_worker])
                worker_addrs.append(worker_addr)
                all_workers.append(worker)
                loaded_wokers.append(worker)
        else:
            if verbose:
                print(f"Using existing worker at {worker_addrs}")
            if not isinstance(worker_addrs, list):
                worker_addrs = [worker_addr]
        call_model_worker = partial(call_worker_func, worker_addrs=worker_addrs)        
    else:
        raise ValueError(f"Engine {engine} not supported, available engines: {ENGINES}")
    
    # wrap the call_model_worker with the model_name and other arguments
    call_model_worker = partial(call_model_worker, model_name=model_name)
    # test local worker connection
    if not completion:
        test_response = call_model_worker(["Hello"], temperature=0, max_tokens=256, timeout=None)
    else:
        test_response = call_model_worker("Hello", temperature=0, max_tokens=256, timeout=None)
    if not test_response:
        print("Error: failed to connect to the worker, exiting...")
        for worker in loaded_wokers:
            cleanup_process(worker)
        sys.exit(1)
    else:
        if verbose:
            print(f"Successfully connected to the workers")
            print("Test prompt: \n", "Hello")
            print("Test response: \n", test_response)
        
    # add cache wrapper
    if use_cache:
        call_model_worker = generation_cache_wrapper(call_model_worker, model_name, cache_dir, overwrite_cache)
    else:
        if verbose:
            print("Cache is disabled")
    call_model_worker = retry_on_failure(call_model_worker, num_retries=max_retry)
    call_model_worker = convert_messages_wrapper(call_model_worker, is_completion=completion)
    set_do_cleanup(True)
    set_verbose(verbose)
    if return_worker:
        return call_model_worker, loaded_wokers
    return call_model_worker


do_cleanup = True
def set_do_cleanup(value):
    global do_cleanup
    do_cleanup = value
    
def kill_process_and_children(pid):
    # check if the process is still alive
    if 'psutil' not in sys.modules:
        # possibly the main process is in the final stage of termination, no need to kill the child processes
        return None
    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        print(f"No process with PID {pid} found.")
        return

    children = parent.children(recursive=True)
    for child in children:
        try:
            child.kill()
            # print(f"Killed child process {child.pid}")
        except psutil.NoSuchProcess:
            print(f"Child process {child.pid} already terminated.")
    
    try:
        parent.kill()
        # print(f"Killed parent process {pid}")
    except psutil.NoSuchProcess:
        print(f"Parent process {pid} already terminated.")
    return True
        
def cleanup_process(worker:Union[ModelWorker, SubprocessMonitor, subprocess.Popen]):
    # os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
    # os.kill(proc.pid, signal.SIGTERM)
    if isinstance(worker, ModelWorker):
        proc = worker.proc.proc
    elif isinstance(worker, SubprocessMonitor):
        proc = worker.proc
    elif isinstance(worker, subprocess.Popen):
        proc = worker
    else:
        raise ValueError(f"Unknown process type {type(proc)}")
    killed = kill_process_and_children(proc.pid)
    if verbose and killed:
        print(f"Model Worker terminated: {worker} ")
    return killed

@atexit.register
def cleanup_all_workers():
    if not do_cleanup:
        return
    for worker in all_workers:
        cleanup_process(worker)
    if all_workers and verbose:
        print("All workers terminated.")
    all_workers.clear()


class LLMEngine:
    
    def __init__(self, verbose=False, num_gpus: int = None, gpu_ids: Union[List[int], str] = None):
        self.workers = []
        self.loaded_model_call_func = {}
        self.loaded_model_engine_map = {}
        import torch
        self.verbose = verbose
        total_gpus = torch.cuda.device_count()
        if gpu_ids:
            assert isinstance(gpu_ids, (list, str)), "passed gpu_ids must be a list or a string"
            if isinstance(gpu_ids, str):
                gpu_ids = [int(gpu_id) for gpu_id in gpu_ids.split(",")]
            assert all(isinstance(gpu_id, int) for gpu_id in gpu_ids), "passed gpu_ids must be a list of integers"
            assert len(gpu_ids) <= total_gpus, f"Error: passed gpu_ids {gpu_ids} is greater than total gpus {total_gpus}"
        else:
            if os.environ.get("CUDA_VISIBLE_DEVICES") is not None:
                gpu_ids = [int(gpu_id) for gpu_id in os.environ.get("CUDA_VISIBLE_DEVICES").split(",")]
            else:
                gpu_ids = list(range(total_gpus))
            if num_gpus is not None:
                num_gpus = int(num_gpus)
                assert num_gpus <= total_gpus, f"Error: passed num_gpus {num_gpus} is greater than total gpus {total_gpus}"
                gpu_ids = gpu_ids[:num_gpus]
        self.gpu_ids = gpu_ids
        self.num_gpus = len(gpu_ids)
        if verbose:
            print(f"LLMEngine initialized with {self.num_gpus} GPUs: {gpu_ids}")
        
    def get_available_gpu_ids(self):
        worker_used_gpu_ids = []
        for worker in self.workers:
            worker_used_gpu_ids.extend(worker.gpu_ids)
        available_gpu_ids = [gpu_id for gpu_id in self.gpu_ids if gpu_id not in worker_used_gpu_ids]
        return available_gpu_ids
    
    def load_model(
        self,
        model_name,
        worker_addrs=None,
        cache_dir=None,
        use_cache=True,
        overwrite_cache=False,
        completion=False,
        num_workers=1,
        num_gpu_per_worker=None,
        dtype="auto",
        quantization=None,
        engine="vllm",
        additional_args=[],
        max_retry=None,
        verbose=None
    ):
        """
        Load a model
        Args:
            model_name: model name
            worker_addrs: worker addresses, if None, launch local workers
            cache_dir: cache directory
            use_cache: use cache or not. Cache is on the hash of the input message.
            overwrite_cache: overwrite cache or not. If True, previous cache will be overwritten.
            completion: use completion or not (use chat by default)
            num_workers: number of workers
            num_gpu_per_worker: number of gpus per worker
            dtype: data type
            engine: engine name
            additional_args: additional arguments for launching the worker (vllm, sglang)
            max_retry: maximum number of retries
            verbose: verbose
        """
        verbose = self.verbose or verbose
        if self.workers:
            print("Warning: previous workers are not cleaned up, please call unload_model() to clean up previous workers")
        self.model_name = model_name
        available_gpu_ids = self.get_available_gpu_ids()
        if engine in ["vllm", "sglang"]:
            num_required_gpus = num_workers * num_gpu_per_worker
            if len(available_gpu_ids) < num_required_gpus:
                print("Error: No available GPU to launch the model worker")
                print("Provided GPU IDs for this LLMEngine class: ", self.gpu_ids)
                print("Used GPU IDs for all workers: ", [worker.gpu_ids for worker in self.workers])
                print("Available GPU IDs: ", available_gpu_ids)
                print("Number of required GPUs: ", num_required_gpus)
                raise ValueError("Not enought available GPU to launch the model worker")
            gpu_ids = available_gpu_ids[:num_required_gpus]
        else:
            num_required_gpus = num_workers
            gpu_ids = None
        
        call_model_worker, model_workers = get_call_worker_func(
            model_name, 
            worker_addrs=worker_addrs, 
            cache_dir=cache_dir, 
            use_cache=use_cache,
            overwrite_cache=overwrite_cache,
            completion=completion, 
            num_workers=num_workers,
            num_gpu_per_worker=num_gpu_per_worker,
            gpu_ids=gpu_ids,
            dtype=dtype,
            quantization=quantization,
            engine=engine,
            additional_args=additional_args,
            max_retry=max_retry,
            verbose=verbose,
            return_worker=True,
        )

        self.workers.extend(model_workers)
        self.loaded_model_call_func[model_name] = call_model_worker
        self.loaded_model_engine_map[model_name] = engine
        return call_model_worker
    
    def call_model(
        self, 
        model_name, 
        messages:Union[List[str], List[dict], str],
        timeout:int=60, 
        conv_system_msg=None, 
        **generate_kwargs
    ):
        """
        Call a model
        Args:
            model_name: model name
            messages: list of messages in openai format
            timeout: timeout
            conv_system_msg: conversation system message
            generate_kwargs: generation arguments
        """
        call_model_worker = self.loaded_model_call_func.get(model_name)
        if call_model_worker is None:
            raise ValueError(f"Model {model_name} not loaded, please call load_model() first")
        try:
            return call_model_worker(messages, timeout=timeout, conv_system_msg=conv_system_msg, **generate_kwargs)
        except MaxRetriesExceededError as e:
            print(e)
            return None
    
    def batch_call_model(
        self,
        model_name,
        batch_messages:List[Union[List[str], List[dict], str]],
        timeout:int=60,
        conv_system_msg=None,
        num_proc=8,
        desc=None,
        disable_openai_batch_api=False,
        **generate_kwargs
    ):
        """
        Batch call a model
        Args:
            model_name: model name
            batch_messages: list of list of messages in openai format or list of strings
            timeout: timeout
            conv_system_msg: conversation system message
            num_proc: number of processes
            generate_kwargs: generation arguments
        """
        supported_batch_api_engines = ["openai", "claude"]
        call_model_worker = self.loaded_model_call_func.get(model_name)
        engine = self.loaded_model_engine_map.get(model_name)
        if engine not in supported_batch_api_engines or disable_openai_batch_api:
            if call_model_worker is None:
                raise ValueError(f"Model {model_name} not loaded, please call load_model() first")
            from functools import partial
            from multiprocessing import Pool
            num_proc = min(num_proc, len(batch_messages))
            call_model_worker = partial(call_model_worker, timeout=timeout, conv_system_msg=conv_system_msg, **generate_kwargs)
            with Pool(num_proc) as p:
                results = list(tqdm(p.imap(call_model_worker, batch_messages), total=len(batch_messages), desc=desc or "LLMEngine Batch Inference"))
        else:
            if engine == "openai":
                print("Using OpenAI batch API")
                from .openai_text import openai_batch_request
                results = openai_batch_request(model_name, batch_messages, conv_system_msg=conv_system_msg, desc=desc, **generate_kwargs)
            elif engine == "claude":
                print("Using Claude batch API")
                from .claude import claude_batch_request
                results = claude_batch_request(model_name, batch_messages, conv_system_msg=conv_system_msg, desc=desc, **generate_kwargs)
            else:
                raise ValueError(f"Engine {engine} not supported for batch API")
                
        return results
    
    def __call__(self, *args, **kwds):
        return self.call_model(*args, **kwds)
                    
    def unload_model(self, model_name=None):
        to_remove_local_workers = []
        to_remove_global_workers = []
        for worker in self.workers:
            if model_name is None or worker.model_name == model_name:
                print(f"Unloading model worker: {worker}")
                cleanup_process(worker) 
                if worker in all_workers:
                    to_remove_global_workers.append(worker)
                if worker in self.workers:
                    to_remove_local_workers.append(worker)
        for worker in to_remove_global_workers:
            all_workers.remove(worker)
        for worker in to_remove_local_workers:
            self.workers.remove(worker)
        
    def __del__(self):
        pass
        
        


from typing import List, Union
from cachetools import LRUCache
from functools import lru_cache
import os
import hashlib
from pathlib import Path

@lru_cache(maxsize=None)
def get_cache_file(model_name_or_path, cache_dir):
    model_name = model_name_or_path.split("/")[-2:]
    model_name = "/".join(model_name)
    if cache_dir is not None:
        cache_file = Path(cache_dir) / f"{model_name}.jsonl"
    else:
        cache_file = Path(os.path.expanduser(f"~/llm_engines/generation_cache/{model_name}.jsonl"))
    if not cache_file.parent.exists():
        cache_file.parent.mkdir(parents=True)
    return cache_file


def get_inputs_hash(inputs, conv_system_msg):
    if isinstance(inputs, str):
        return hashlib.md5(inputs.encode()).hexdigest()
    if conv_system_msg:
        return hashlib.md5((conv_system_msg + "".join(inputs)).encode()).hexdigest()
    return hashlib.md5("".join(inputs).encode()).hexdigest()


import fire
import random
import os
from typing import List, Union

class Cli:
    def __init__(self):
        pass

    def serve(self,
        model_name: str,
        engine: str="vllm",
        num_gpus: int=1,
        gpu_ids: List[int]=None,
        dtype: str="auto",
        quantization: str=None,
        port: Union[int,None] =None,
        host: str="127.0.0.1",
    ) -> str:
        from . import set_do_cleanup
        set_do_cleanup(False)
        assert engine in ["vllm", "sglang"]
        if engine == "vllm":
            from .vllm import launch_vllm_worker
            launch_worker_func = launch_vllm_worker
        elif engine == "sglang":
            from .sglang import launch_sglang_worker
            launch_worker_func = launch_sglang_worker
        if port is None:
            print("Warning: port not provided, using random port in range 31000-32000")
            port = random.randint(31000, 32000)
        if isinstance(gpu_ids, int):
            gpu_ids = [gpu_ids]
        elif isinstance(gpu_ids, str):
            gpu_ids = [int(gpu_id) for gpu_id in gpu_ids.split(",")]
        else:
            gpu_ids = None
        workder_addr, proc = launch_worker_func(
            model_name=model_name,
            num_gpus=num_gpus,
            gpu_ids=gpu_ids,
            dtype=dtype,
            quantization=quantization,
            port=port,
            host=host
        )
        
    def call_worker(
        self,
        model_name: str,
        worker_addr: str,
        prompt: str,
        engine: str="vllm",
        temperature: float=0.0,
        top_p: float=1.0,
        max_tokens: int=None,
        timeout: int=60,
    ) -> str:
        from . import get_call_worker_func
        call_worker_func = get_call_worker_func(
            model_name=model_name,
            worker_addrs=[worker_addr],
            engine=engine,
            use_cache=True,
            overwrite_cache=False
        )
        response = call_worker_func([prompt], temperature=temperature, top_p=top_p, max_tokens=max_tokens, timeout=timeout)
        return response
    
    def clean_workers(
        self,
        engine: str="vllm",
    ):
        print("Note: This will kill all workers for the specified engine")
        if engine == "vllm":
            os.system("pkill -f vllm.entrypoints.openai.api_server")
        elif engine == "sglang":
            os.system("pkill -f sglang.launch_server")
        else:
            raise ValueError(f"Engine {engine} not supported")
        print("Workers cleaned")
        
def main():
    fire.Fire(Cli)
    
if __name__ == "__main__":
    main()
        
        
"""
llm-engines serve "meta-llama/Meta-Llama-3-8B-Instruct" --engine vllm --num-gpus 1 --gpu-ids 2 --dtype auto 
llm-engines call-worker "meta-llama/Meta-Llama-3-8B-Instruct" "http://127.0.0.1:31845" "Hello"
llm-engines clean-workers --engine vllm

"""


import subprocess
import time
import os
import signal
import os
import signal
import json
import hashlib
import traceback
import threading
import openai
import inspect
from pathlib import Path
from typing import Union, List
from typing import List
from transformers import AutoTokenizer
from functools import partial
from .cache import load_cache, get_inputs_hash, get_cache_file

default_gen_params = {
    "temperature": 0.0,
    "max_tokens": None,
    "top_p": 1.0,
    "timeout": 600,
}
class SubprocessMonitor:
    def _monitor(self):
        while True:
            if self.proc.poll() is not None:
                print("Subprocess has exited with code", self.proc.returncode)
                os.kill(os.getpid(), signal.SIGTERM)  # Exit the main process
                break
            time.sleep(5)
            
    def __init__(self, command, **kwargs):
        print("Launching subprocess with command:\n", " ".join(command))
        self.proc = subprocess.Popen(command, **kwargs)
        # self.monitor_thread = threading.Thread(target=self._monitor)
        # self.monitor_thread.start()
    
class ChatTokenizer:
    def __init__(self, model_name):
        
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
        self.system_message = None
        try:
            self.max_length = self.tokenizer.model_max_length
        except AttributeError:
            self.max_length = 4096
        if not isinstance(self.max_length, int):
            self.max_length = 4096
        if self.max_length > 1e6:
            self.max_length = 1e6
            
        if self.tokenizer.chat_template:
            self.apply_chat_template = self.apply_chat_template_default
            print("Using hugging face chat template for model", model_name)
            self.chat_template_source = "huggingface"
        else:
            self.apply_chat_template = None
            self.chat_template_source = None
        print("Example prompt: \n", self.example_prompt())
        
    def apply_chat_template_default(
        self, 
        messages:List[str],
        add_generation_prompt:bool=True,
        chat_template:str=None
    ):
        prompt = self.tokenizer.apply_chat_template(
            messages, 
            add_generation_prompt=add_generation_prompt,
            tokenize=False,
            chat_template=chat_template,
        )
        return prompt
    
    def example_prompt(self):
        if not self.apply_chat_template:
            return "Chat template not available for this model"
        else:
            example_messages = [
                {"role": "user", "content": "Hello"},
                {"role": "assistant", "content": "Hi"},
                {"role": "user", "content": "How are you?"},
                {"role": "assistant", "content": "I'm good, how about you?"},
            ]
            return self.apply_chat_template(example_messages)
    
    def __call__(self, messages:List[str], **kwargs):
        if not self.apply_chat_template:
            raise NotImplementedError("Chat template not available for this model")
        return self.apply_chat_template(messages, **kwargs)


def convert_messages(messages:Union[List[str], List[dict], str]):
    """
    Convert messages to the format expected by the model
    """
    if all(isinstance(item, dict) for item in messages):
        assert all("content" in item for item in messages), "content key not found in messages"
        assert all("role" in item for item in messages), "role key not found in messages"
        if messages[0]["role"] == "system":
            conv_system_msg = messages[0]["content"]
            messages = messages[1:]
        else:
            conv_system_msg = None
        new_messages = []
        for i, message in enumerate(messages):
            if i % 2 == 0:
                assert message["role"] == "user", "User message must be at even index"
            else:
                assert message["role"] == "assistant", "Assistant message must be at odd index"
            new_messages.append(message["content"])
        return new_messages, conv_system_msg
    else:
        if isinstance(messages, str):
            messages = [messages]
        assert all(isinstance(item, str) for item in messages)
        return messages, None

def _convert_messages_wrapper(messages:Union[List[str], List[dict], str], call_model_worker, is_completion=False, **generate_kwargs):
    if not is_completion:
        messages, conv_system_msg = convert_messages(messages)
        generate_kwargs["conv_system_msg"] = conv_system_msg
    else:
        assert isinstance(messages, str), "Completion model only accepts a single string input"
    # add default generation parameters
    for key, value in default_gen_params.items():
        if key not in generate_kwargs:
            generate_kwargs[key] = value
    return call_model_worker(messages, **generate_kwargs)

def convert_messages_wrapper(call_model_worker, is_completion=False):
    return partial(_convert_messages_wrapper, call_model_worker=call_model_worker, is_completion=is_completion)

def _generation_cache_wrapper(inputs: Union[str, List[str]], call_model_worker, model_name, cache_dir=None, overwrite_cache=False, **generate_kwargs):
    cache_dict = load_cache(model_name, cache_dir)
    
    conv_system_msg = generate_kwargs.get("conv_system_msg", "")
    inputs_hash = get_inputs_hash(inputs, conv_system_msg)
    
    if not overwrite_cache:
        cached_value = cache_dict[inputs_hash]
        if cached_value:
            return cached_value["output"]
    
    generated_text = call_model_worker(inputs, **generate_kwargs)
    cache_item = {
        "input": inputs,
        "output": generated_text,
        "model_name": model_name,
        'tstamp': time.time(),
        "time": time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
        "generate_kwargs": generate_kwargs
    }
    
    # cache_dict[inputs_hash] = cache_item
    
    cache_file = get_cache_file(model_name, cache_dir)
    with open(cache_file, "a+") as f:
        f.write(json.dumps({inputs_hash: cache_item}) + "\n")
    
    return generated_text

def generation_cache_wrapper(call_model_worker, model_name, cache_dir=None, overwrite_cache=False):
    print(f"Using efficient multi-level cache for model {model_name}")
    if cache_dir is None:
        env_cache_dir = os.getenv("LLM_ENGINES_CACHE_DIR")
        if env_cache_dir:
            cache_dir = Path(env_cache_dir)
        else:
            cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache"))
    print(f"Cache directory: {cache_dir}")
    load_cache(model_name, cache_dir) # preload cache
    
    return partial(_generation_cache_wrapper, call_model_worker=call_model_worker, model_name=model_name, cache_dir=cache_dir, overwrite_cache=overwrite_cache)
class MaxRetriesExceededError(Exception):
    pass

short_error_instances = [
    openai.BadRequestError,
]

def _retry_on_failure(*args, call_model_worker=None, num_retries=5, **kwargs):
    if not call_model_worker:
        raise ValueError("call_model_worker is required")
    try:
        return call_model_worker(*args, **kwargs)
    except Exception as e:
        if not num_retries:
            if any(isinstance(e, error_instance) for error_instance in short_error_instances):
                print(str(e))
            else:
                print(traceback.format_exc())
            raise MaxRetriesExceededError(f"Max retries exceeded for call_model_worker (num_retries={num_retries})")
        for i in range(num_retries):
            try:
                return call_model_worker(*args, **kwargs)
            except Exception as e:
                print("Error in call_model_worker, retrying... (Error: {})".format(e))
                time.sleep(1)
                if i >= num_retries - 1 and not isinstance(e, TimeoutError):
                    if any(isinstance(e, error_instance) for error_instance in short_error_instances):
                        print(str(e))
                    else:
                        # format dump of the last error and
                        print(traceback.format_exc())
        raise MaxRetriesExceededError(f"Max retries exceeded for call_model_worker (num_retries={num_retries})")
    
def retry_on_failure(call_model_worker, num_retries=5):
    return partial(_retry_on_failure, call_model_worker=call_model_worker, num_retries=num_retries)
    return wrapper

def timeout_handler(signum, frame):
    raise TimeoutError("Function call timed out")

def with_timeout(timeout):
    def decorator(func):
        def wrapper(*args, **kwargs):
            result = [TimeoutError(f"Function call timed out (timeout={timeout})")]
            stop_event = threading.Event()

            def target():
                try:
                    result[0] = func(*args, **kwargs)
                except Exception as e:
                    result[0] = e

            thread = threading.Thread(target=target)
            thread.start()
            thread.join(timeout)
            if thread.is_alive():
                stop_event.set()
                raise TimeoutError(f"Function call timed out (timeout={timeout})")
            if isinstance(result[0], Exception):
                raise result[0]
            return result[0]
        return wrapper
    return decorator

def get_function_arg_names(func):
    signature = inspect.signature(func)
    parameters = signature.parameters
    
    arg_names = []
    kwarg_names = []
    
    for name, param in parameters.items():
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            arg_names.append(f"*{name}")
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            kwarg_names.append(f"**{name}")
        else:
            arg_names.append(name)
    
    return arg_names, kwarg_names


import os
import json
import hashlib
import time
from datetime import datetime
from openai import OpenAI
from typing import List, Union
from pathlib import Path
from tqdm import tqdm

batch_submission_status_file = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "openai_batch_cache" / "batch_submission_status.json"
batch_submission_status_file.parent.mkdir(parents=True, exist_ok=True)
# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_openai(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to openai format
    new_messages = []
    if conv_system_msg:
        new_messages.append({"role": "system", "content": conv_system_msg})
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})
    # initialize openai client
    client = OpenAI()
    # call openai
    completion = client.chat.completions.create(
        model=model_name,
        messages=new_messages,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.message.content for c in completion.choices]
        else:
            return completion.choices[0].message.content
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        return generate_stream()

def call_worker_openai_completion(prompt:str, model_name, timeout:int=60, **generate_kwargs) -> str:
    # initialize openai client
    client = OpenAI()
    # call openai
    completion = client.completions.create(
        model=model_name,
        prompt=prompt,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.text for c in completion.choices]
        else:
            return completion.choices[0].text
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].text is not None:
                    yield chunk.choices[0].text
        return generate_stream()
    
def save_batch_file(batch_messages:List[Union[str, dict]], model:str, batch_name:str=None, cache_dir=None, custom_ids=None, **generate_kwargs):
    if isinstance(batch_messages[0], str):
        batch_messages = [{"role": "user", "content": message} for message in batch_messages]

    if cache_dir is None:
        cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "openai_batch_cache"
    else:
        cache_dir = Path(cache_dir) / "openai_batch_cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    if custom_ids is None:
        custom_ids = [f"request-{i+1}" for i in range(len(batch_messages))]
    assert len(custom_ids) == len(batch_messages)
    if batch_name is None:
        # batch_name = f"{model}_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        hash_str = "".join([str(message) for message in batch_messages])
        batch_name = f"{model}_{hashlib.md5(hash_str.encode()).hexdigest()}"
    file_path = cache_dir / f"{batch_name}.jsonl"
    with open(file_path, "w") as f:
        for custom_id, message in zip(custom_ids, batch_messages):
            if isinstance(message, str):
                message = {"role": "user", "content": message}
            f.write(json.dumps({"custom_id": custom_id, "method": "POST", "url": "/v1/chat/completions", "body": {"model": model, "messages": message, **generate_kwargs}}, ensure_ascii=False) + "\n")
    return file_path

def submit_batch_file(batch_file:str, output_path:str=None, project_name:str=None, description:str=None):
    """
    Submit a batch of queries to OpenAI API
    Args:
        batch_file: str, path to the batch jsonl file
        output_path: str, path to the output file; if not provided, will be saved in the same directory as the batch_file, ending with ".batch_results.jsonl"
        project_name: str, project name, default to "vapo", keyword to filter batches
        description: str, description of the batch, default to the batch_file name, can also be used to filter batches
    Returns:
        batch_result_id: str, the id of the batch submission
    """
    if isinstance(batch_file, str):
        batch_file = Path(batch_file)
        
    # internally maintain a batch submission status json
    print("Loading batch submission status from", batch_submission_status_file)
    if batch_submission_status_file.exists():
        try:
            with open(batch_submission_status_file, "r") as f:
                batch_submission_status = json.load(f)
        except:
            batch_submission_status = {}
    else:
        batch_submission_status = {}
        
    
    client = OpenAI()
    batch_input_file = client.files.create(
        file=open(batch_file, "rb"),
        purpose="batch"
    )
    if not project_name:
        project_name = "llm_engines"
    description = description if description is not None else batch_file.stem
    output_path = output_path if output_path is not None else batch_file.parent / f"{description}.batch_results.jsonl"
    batch_input_file_id = batch_input_file.id
    
    batch_file_hash = hashlib.md5(batch_file.read_bytes()).hexdigest()
    batch_file_size = batch_file.stat().st_size
    
    for key, value in batch_submission_status.items():
        value_input_file_metadata = value["input_path_metadata"]
        if not os.path.samefile(value["input_path"], batch_file):
            # print(f"Batch {key} has a different input file. need to resubmit.")
            continue
        if batch_file_size != value_input_file_metadata["size"]:
            # print(f"Batch {key} has a newer version of the input file. need to resubmit.")
            continue
        if batch_file_hash != value_input_file_metadata["hash"]:
            # print(f"Batch {key} has a different input hash. need to resubmit.")
            continue
        print(f"Batch {key} already submitted. Skipping submission.")
        return key
        
    completion_window = "24h"
    endpoint = "/v1/chat/completions"
    batch_result = client.batches.create(
        input_file_id=batch_input_file_id,
        endpoint=endpoint,
        completion_window=completion_window,
        metadata={
            "project": project_name,
            "description": description,
            "input_path": str(batch_file),
            "output_path": str(output_path)
        }
    )
    print(f"Batch {batch_result.id} submitted")
    
    
    # time should be in the current timezone, in the format like 2022-01-01T00:00:00
    batch_submission_status[batch_result.id] = {    
        "project": project_name,
        "description": description,
        "endpoint": endpoint,
        "completion_window": completion_window,
        "input_path": str(batch_file),
        "input_path_metadata": {
            "hash": hashlib.md5(batch_file.read_bytes()).hexdigest(),
            "size": batch_file.stat().st_size
        },
        "output_path": str(output_path),
        "batch_input_file_id": batch_input_file_id,
        "batch_result_id": batch_result.id,
        "status": "submitted",
        "timeline": {
            "submitted": str(datetime.now()),
            "completed": None,
            "failed": None,
            "downloaded": None
        },
        "last_updated": str(datetime.now()),
        "openai_batch_metadata": batch_result.to_dict()
    }
    with open(batch_submission_status_file, "w") as f:
        json.dump(batch_submission_status, f, indent=4)
    return batch_result.id

def check_batch_status(batch_result_id, overwrite:bool=False):
     # internally maintain a batch submission status json
    if batch_submission_status_file.exists():
        with open(batch_submission_status_file, "r") as f:
            batch_submission_status = json.load(f)
    else:
        batch_submission_status = {}
    
    client = OpenAI()
    batch = client.batches.retrieve(batch_result_id)
    batch_id = batch.id
    batch_status = batch.status
    batch_desc = batch.metadata["description"] if batch.metadata is not None and "description" in batch.metadata else ""
    batch_project_name = batch.metadata["project"] if batch.metadata is not None and "project" in batch.metadata else ""
    if batch.metadata is not None and "output_path" in batch.metadata:
        batch_output_path = batch.metadata["output_path"]
    else:
        batch_output_path = f"./batch_results/{batch_id}.batch_results.jsonl"
        batch_submission_status[batch_id]["output_path"] = batch_output_path
    # print(f"{batch_id: <20} {batch_status: <20} {batch_project_name: <20} {batch_desc: <20}")
    if batch_status == "completed":
        print(f"Downloading output file for batch {batch_id}")
        content = client.files.content(batch.output_file_id)
        output_path = batch_output_path
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        if output_path.exists() and not overwrite:
            print(f"File {output_path} already exists. Skipping writing to file.")
        else:
            if output_path.exists() and overwrite:
                print(f"Overwriting file {output_path}")
            content.write_to_file(output_path)
            print(f"Output file written to {output_path}")
        batch_submission_status[batch_id]["status"] = "completed"
        batch_submission_status[batch_id]["timeline"]["completed"] = str(datetime.now())
        batch_submission_status[batch_id]["timeline"]["downloaded"] = str(datetime.now())
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["openai_batch_metadata"].update(batch.to_dict())
    elif batch_status == "failed":
        print(f"Batch {batch_id} failed.")
        batch_submission_status[batch_id]["status"] = "failed"
        batch_submission_status[batch_id]["timeline"]["failed"] = str(datetime.now())
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["openai_batch_metadata"].update(batch.to_dict())
    else:
        batch_submission_status[batch_id]["status"] = batch_status
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["openai_batch_metadata"].update(batch.to_dict())
    with open(batch_submission_status_file, "w") as f:
        json.dump(batch_submission_status, f, indent=4)
    return batch_submission_status[batch_id]

def openai_batch_request(
    model_name:str,
    batch_messages:List[Union[str, List[str], List[dict]]],
    conv_system_msg:str=None,
    desc:str=None,
    **generate_kwargs
):
    if isinstance(batch_messages[0], str):
        batch_messages = [[{"role": "user", "content": message}] for message in batch_messages]
    elif isinstance(batch_messages[0], list):
        if isinstance(batch_messages[0][0], str):
            batch_messages = [[{
                "role": "user" if i % 2 == 0 else "assistant",
                "content": message
            } for i, message in enumerate(messages)] for messages in batch_messages]
        elif isinstance(batch_messages[0][0], dict):
            assert all("role" in message for message in batch_messages[0]), "Error: role key not found in the message"
            assert all("content" in message for message in batch_messages[0]), "Error: content key not found in the message"
        else:
            raise ValueError("Error: unknown message format")
    else:
        raise ValueError("Error: unknown message format")
    if conv_system_msg:
        batch_messages = [[{"role": "system", "content": conv_system_msg}] + messages for messages in batch_messages]
    if "stream" in generate_kwargs:
        generate_kwargs.pop("stream")
    batch_file = save_batch_file(batch_messages, model_name, **generate_kwargs)
    batch_result_id = submit_batch_file(batch_file)
    num_total = len(batch_messages)
    tqdm_bar = tqdm(total=num_total, desc=desc or "LLMEngine Batch Inference")
    while True:
        batch_status = check_batch_status(batch_result_id)
        num_completed = batch_status["openai_batch_metadata"]['request_counts']['completed']
        num_total = batch_status["openai_batch_metadata"]['request_counts']['total']
        num_failed = batch_status["openai_batch_metadata"]['request_counts']['failed']
        tqdm_bar.n = num_completed
        tqdm_bar.total = num_total
        tqdm_bar.set_postfix(completed=num_completed, total=num_total, failed=num_failed)
        if batch_status["status"] == "completed":
            tqdm_bar.close()
            break
        elif batch_status["status"] == "finalizing":
            tqdm_bar.desc = "Finalizing"
            tqdm_bar.refresh()
        elif batch_status["status"] == "validating":
            tqdm_bar.desc = "Validating"
            tqdm_bar.refresh()
        elif batch_status["status"] == "failed":
            tqdm_bar.close()
            print("Batch failed")
            break
        elif batch_status["status"] == "in_progress":
            tqdm_bar.desc = "In Progress"
            tqdm_bar.refresh()
        elif batch_status["status"] == "cancelled":
            tqdm_bar.close()
            print("Batch cancelled")
            break
        elif batch_status["status"] == "expired":
            tqdm_bar.close()
            print("Batch expired")
            break
        elif batch_status["status"] == "cancelling":
            tqdm_bar.desc = "Cancelling"
            tqdm_bar.refresh()
        else:
            tqdm_bar.desc = batch_status["status"]
            tqdm_bar.refresh()
        time.sleep(10)
        
    if batch_status["status"] == "completed":
        output_path = batch_status["output_path"]
        results = []
        with open(output_path, "r") as f:
            results = [json.loads(line) for line in f.readlines()]
        all_completions = [[choice['message']['content'] for choice in x['response']['body']['choices']] for x in results]
        if all(len(x) == 1 for x in all_completions):
            all_completions = [x[0] for x in all_completions]
        results = all_completions
    else:
        results = None
    return results

if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_openai(["Hello"], "gpt-3.5-turbo"))
    ic(call_worker_openai_completion("Hello", "gpt-3.5-turbo-instruct"))


from .cache_sqlite3 import load_cache as load_sqlite3_cache
from .cache_lru import load_cache as load_lru_cache
from .cache_dict import load_cache as load_dict_cache
from .cache_utils import get_cache_file, get_inputs_hash

load_cache = load_sqlite3_cache


import anthropic
import os
import requests
import json
import hashlib
import time
from typing import List, Union
from anthropic import NOT_GIVEN
from datetime import datetime
from pathlib import Path
from tqdm import tqdm
from .utils import with_timeout
batch_submission_status_file = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "claude_batch_cache" / "batch_submission_status.json"
batch_submission_status_file.parent.mkdir(parents=True, exist_ok=True)

# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_claude(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to mistral format
    client = anthropic.Anthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))
    new_messages = []
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})
             
    generate_kwargs.pop("n", None) # claude does not have n
    if not generate_kwargs.get("max_tokens", None):
        generate_kwargs["max_tokens"] = 1024
    stream = generate_kwargs.pop("stream", False)
    @with_timeout(timeout)
    def get_response():
        completion = client.messages.create(
            model=model_name,
            messages=new_messages,
            system=conv_system_msg if conv_system_msg else NOT_GIVEN,
            timeout=timeout,
            **generate_kwargs,
        )
        if len(completion.content) > 1:
            return [c.text for c in completion.content]
        else:
            return completion.content[0].text
        
    @with_timeout(timeout)
    def stream_response():
        with client.messages.stream(
            model=model_name,
            messages=new_messages,
            system=conv_system_msg if conv_system_msg else NOT_GIVEN,
            timeout=timeout,
            **generate_kwargs,
        ) as stream:
            for text in stream.text_stream:
                yield text
                
    if not stream:
        return get_response()
    else:
        return stream_response()


def save_batch_file(batch_messages:List[Union[str, dict]], model:str, batch_name:str=None, cache_dir=None, custom_ids=None, **generate_kwargs):
    if isinstance(batch_messages[0], str):
        batch_messages = [{"role": "user", "content": message} for message in batch_messages]

    if cache_dir is None:
        cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "openai_batch_cache"
    else:
        cache_dir = Path(cache_dir) / "openai_batch_cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    if custom_ids is None:
        custom_ids = [f"request-{i+1}" for i in range(len(batch_messages))]
    assert len(custom_ids) == len(batch_messages)
    if batch_name is None:
        # batch_name = f"{model}_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        hash_str = "".join([str(message) for message in batch_messages])
        batch_name = f"{model}_{hashlib.md5(hash_str.encode()).hexdigest()}"
    file_path = cache_dir / f"{batch_name}.jsonl"
    with open(file_path, "w") as f:
        for custom_id, message in zip(custom_ids, batch_messages):
            if isinstance(message, str):
                message = {"role": "user", "content": message}
            f.write(json.dumps({"custom_id": custom_id, "params": {"model": model, "messages": message, **generate_kwargs}}, ensure_ascii=False) + "\n")
    return file_path

def submit_batch_file(batch_file:str, output_path:str=None, project_name:str=None, description:str=None):
    """
    Submit a batch of queries to OpenAI API
    Args:
        batch_file: str, path to the batch jsonl file
        output_path: str, path to the output file; if not provided, will be saved in the same directory as the batch_file, ending with ".batch_results.jsonl"
        project_name: str, project name, default to "vapo", keyword to filter batches
        description: str, description of the batch, default to the batch_file name, can also be used to filter batches
    Returns:
        batch_result_id: str, the id of the batch submission
    """
    if isinstance(batch_file, str):
        batch_file = Path(batch_file)
        
    # internally maintain a batch submission status json
    print("Loading batch submission status from", batch_submission_status_file)
    if batch_submission_status_file.exists():
        try:
            with open(batch_submission_status_file, "r") as f:
                batch_submission_status = json.load(f)
        except:
            batch_submission_status = {}
    else:
        batch_submission_status = {}
        
    
    client = anthropic.Anthropic()
    with open(batch_file, "r") as f:
        batch_inputs = [json.loads(line) for line in f.readlines()]
    if not project_name:
        project_name = "llm_engines"
    description = description if description is not None else batch_file.stem
    output_path = output_path if output_path is not None else batch_file.parent / f"{description}.batch_results.jsonl"
    
    batch_file_hash = hashlib.md5(batch_file.read_bytes()).hexdigest()
    batch_file_size = batch_file.stat().st_size
    
    for key, value in batch_submission_status.items():
        value_input_file_metadata = value["input_path_metadata"]
        if not os.path.samefile(value["input_path"], batch_file):
            # print(f"Batch {key} has a different input file. need to resubmit.")
            continue
        if batch_file_size != value_input_file_metadata["size"]:
            # print(f"Batch {key} has a newer version of the input file. need to resubmit.")
            continue
        if batch_file_hash != value_input_file_metadata["hash"]:
            # print(f"Batch {key} has a different input hash. need to resubmit.")
            continue
        print(f"Batch {key} already submitted. Skipping submission.")
        return key
    
    
    batch_result = client.beta.messages.batches.create(
        requests=batch_inputs,
    )
    print(f"Batch {batch_result.id} submitted")
    
    claude_batch_metadata = batch_result.to_dict()
    for key in claude_batch_metadata:
        if isinstance(claude_batch_metadata[key], datetime):
            claude_batch_metadata[key] = str(claude_batch_metadata[key])
    # time should be in the current timezone, in the format like 2022-01-01T00:00:00
    batch_submission_status[batch_result.id] = {    
        "project": project_name,
        "description": description,
        "endpoint": None,
        "completion_window": None,
        "input_path": str(batch_file),
        "input_path_metadata": {
            "hash": hashlib.md5(batch_file.read_bytes()).hexdigest(),
            "size": batch_file.stat().st_size,
            "total": len(batch_inputs)
        },
        "output_path": str(output_path),
        "batch_input_file_id": None,
        "batch_result_id": batch_result.id,
        "status": "submitted",
        "timeline": {
            "submitted": str(datetime.now()),
            "completed": None,
            "failed": None,
            "downloaded": None,
        },
        "last_updated": str(datetime.now()),
        "claude_batch_metadata": claude_batch_metadata
    }
    with open(batch_submission_status_file, "w") as f:
        json.dump(batch_submission_status, f, indent=4)
    return batch_result.id

def check_batch_status(batch_result_id, overwrite:bool=False):
     # internally maintain a batch submission status json
    if batch_submission_status_file.exists():
        with open(batch_submission_status_file, "r") as f:
            batch_submission_status = json.load(f)
    else:
        batch_submission_status = {}
    
    client = anthropic.Anthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))
    batch = client.beta.messages.batches.retrieve(batch_result_id)
    batch_id = batch.id
    batch_status = batch.processing_status
    batch_desc = batch_submission_status[batch_id]["description"] if batch_id in batch_submission_status else ""
    batch_project_name = batch_submission_status[batch_id]["project"] if batch_id in batch_submission_status else ""
    if batch_submission_status[batch_id]["output_path"]:
        batch_output_path = batch_submission_status[batch_id]["output_path"]
    else:
        batch_output_path = f"./batch_results/{batch_id}.batch_results.jsonl"
        batch_submission_status[batch_id]["output_path"] = batch_output_path
    # print(f"{batch_id: <20} {batch_status: <20} {batch_project_name: <20} {batch_desc: <20}")
    claude_batch_metadata = batch.to_dict()
    for key in claude_batch_metadata:
        if isinstance(claude_batch_metadata[key], datetime):
            claude_batch_metadata[key] = str(claude_batch_metadata[key])
    if batch_status == "ended":
        output_path = batch_output_path
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        if output_path.exists() and not overwrite:
            print(f"File {output_path} already exists. Skipping writing to file.")
        else:
            if output_path.exists() and overwrite:
                print(f"Overwriting file {output_path}")
                
            # retrieve the results
            print(f"Downloading output file for batch {batch_id}")
            results = client.beta.messages.batches.results(batch_id)
            # save url to file via web requests, download the results_url file, and save to output_path
            with open(output_path, "wb") as f:
                for result in results:
                    f.write(json.dumps(result.to_dict(), ensure_ascii=False).encode() + b"\n")
            print(f"Output file written to {output_path}")
        with open(output_path, "r") as f:
            results = [json.loads(line) for line in f.readlines()]
        all_result_types = [x['result']['type'] for x in results]
        if all(all_result_types) == "errored" or all(all_result_types) == "error":
            batch_status = "errored"
            batch_submission_status[batch_id]["status"] = "errored"
            batch_submission_status[batch_id]["timeline"]["failed"] = str(datetime.now())
        elif all(all_result_types) == "canceled":
            batch_status = "canceled"
            batch_submission_status[batch_id]["status"] = "canceled"
            batch_submission_status[batch_id]["timeline"]["canceled"] = str(datetime.now())
        elif all(all_result_types) == "expired":
            batch_status = "expired"
            batch_submission_status[batch_id]["status"] = "expired"
            batch_submission_status[batch_id]["timeline"]["expired"] = str(datetime.now())
        else:
            batch_submission_status[batch_id]["status"] = "completed"
        batch_submission_status[batch_id]["timeline"]["completed"] = str(datetime.now())
        batch_submission_status[batch_id]["timeline"]["downloaded"] = str(datetime.now())
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["claude_batch_metadata"].update(claude_batch_metadata)
    else:
        batch_submission_status[batch_id]["status"] = batch_status
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["claude_batch_metadata"].update(claude_batch_metadata)
    with open(batch_submission_status_file, "w") as f:
        json.dump(batch_submission_status, f, indent=4)
    return batch_submission_status[batch_id]

def claude_batch_request(
    model_name:str,
    batch_messages:List[Union[str, List[str], List[dict]]],
    conv_system_msg:str=None,
    desc:str=None,
    **generate_kwargs
):
    if isinstance(batch_messages[0], str):
        batch_messages = [[{"role": "user", "content": message}] for message in batch_messages]
    elif isinstance(batch_messages[0], list):
        if isinstance(batch_messages[0][0], str):
            batch_messages = [[{
                "role": "user" if i % 2 == 0 else "assistant",
                "content": message
            } for i, message in enumerate(messages)] for messages in batch_messages]
        elif isinstance(batch_messages[0][0], dict):
            assert all("role" in message for message in batch_messages[0]), "Error: role key not found in the message"
            assert all("content" in message for message in batch_messages[0]), "Error: content key not found in the message"
        else:
            raise ValueError("Error: unknown message format")
    else:
        raise ValueError("Error: unknown message format")
    if conv_system_msg:
        batch_messages = [[{"role": "system", "content": conv_system_msg}] + messages for messages in batch_messages]
    if "stream" in generate_kwargs:
        generate_kwargs.pop("stream")
    batch_file = save_batch_file(batch_messages, model_name, **generate_kwargs)
    batch_result_id = submit_batch_file(batch_file)
    num_total = len(batch_messages)
    tqdm_bar = tqdm(total=num_total, desc=desc or "LLMEngine Batch Inference")
    while True:
        batch_status = check_batch_status(batch_result_id)
        num_succeeded = batch_status["claude_batch_metadata"]['request_counts']['succeeded']
        num_processing = batch_status["claude_batch_metadata"]['request_counts']['processing']
        num_errored = batch_status["claude_batch_metadata"]['request_counts']['errored']
        num_expired = batch_status["claude_batch_metadata"]['request_counts']['expired']
        num_canceled = batch_status["claude_batch_metadata"]['request_counts']['canceled']
        num_total = num_succeeded + num_processing + num_errored + num_expired + num_canceled
        assert num_total == len(batch_messages), f"Error: total number of requests {num_total} does not match the number of requests {len(batch_messages)}"
        tqdm_bar.n = num_succeeded + num_errored + num_expired + num_canceled
        tqdm_bar.total = num_total
        # tqdm_bar.set_postfix(completed=num_completed, total=num_total, failed=num_failed)
        tqdm_bar.set_postfix(
            completed=num_succeeded,
            processing=num_processing,
            errored=num_errored,
            expired=num_expired,
            canceled=num_canceled
        )
        if batch_status["status"] in ["completed", "errored", "expired", "canceled"]:
            tqdm_bar.close()
            break
        elif batch_status["status"] == "in_progress":
            tqdm_bar.desc = "In Progress"
            tqdm_bar.refresh()
        else:
            tqdm_bar.desc = batch_status["status"]
            tqdm_bar.refresh()
        time.sleep(10)
        
    if batch_status["status"] == "completed":
        output_path = batch_status["output_path"]
        results = []
        with open(output_path, "r") as f:
            raw_results = [json.loads(line) for line in f.readlines()]
        results = []
        for item in raw_results:
            if item["result"]["type"] == "succeeded":
                results.append(item["result"]['message']["content"][0]['text'])
            else:
                results.append(None)
        print("Batch requests status counts:", batch_status["claude_batch_metadata"]['request_counts'])
    else:
        print(f"Error: batch {batch_result_id} failed with status {batch_status['status']}")
        results = None
    return results


if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_claude(["Hello", "Hi, I am claude", "What did I ask in the last response?"], "claude-3-opus-20240229"))